#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#define MIN2(y,z)        ((y)<(z))?(y):(z)
#define MIN3(x,y,z)      MIN2((x),MIN2((y),(z)))
#define MIN4(w,x,y,z)    MIN2((w),MIN3((x),(y),(z)))

// Newick: (((One:0.2,Two:0.3):0.3,(Three:0.5,Four:0.3):0.2):0.3,Five:0.7):0.0;

// Takes a tree in newick format, builds an internal "tree" structure
// generates calls to other programs with correct weights


typedef struct tree_node {
  struct tree_node* lc;
  struct tree_node* rc;
  struct tree_node* par;
  char* seqname; 
  int leafs;
  float distup;
} node;


int numleafs = 0;
int missing_dist = 0;
char** outnames;
float* outdists;  
//int numouts;


void do_postorder(node* myroot);


void init_consts() {
  outnames = (char**) malloc (sizeof(char*) * numleafs);
  outdists = (float*) malloc (sizeof(float) * numleafs);  
}


node* mk_node() {
  node* res = (node*)malloc(sizeof(node));
  res->leafs = 0;
  res->lc = res->rc = res->par = 0;
  res->seqname = 0;
  res->distup = -1.0;
  return res;
}

void parse_error(char* str, int m, char* symb) {
  fprintf(stderr, "Error in input %s\n", str+m);
  fprintf(stderr, "Was expecting %s\n", symb);
  exit(1);
}

int mystrchr(char* str, int ch) {
  int k = (int)strchr(str,ch);
  return (k)? k : 9999999;
}

int read_name (char* str, node* myroot) {
  int n = (int) MIN3(mystrchr(str, ','), mystrchr(str,':'), 
		     mystrchr(str,')'));

  n = n - (int)str;
  //  fprintf(stderr, "n=%d\n", n);
  if (n < 0) {
    parse_error(str, 0, "%s{,:)}");
  }

  numleafs++;
  myroot->seqname = (char*) malloc(n+1);
  strncpy(myroot->seqname, str, n);
  myroot->seqname[n] = 0;
  return n;
}

int read_dist (char* str, node* myroot) {
  float d;
  int n;
  if (!sscanf(str, "%f%n", &d, &n))
    parse_error(str, 0, "%f");
  myroot->distup = d;
  return n;
}

int parse_tree(char* tree_str, node* myroot) {
  int moved = 0;
  //  fprintf(stderr, "Proc %s\n", tree_str);
  if (tree_str[0] != '(') {
    myroot->leafs = 1;
    moved += read_name(tree_str, myroot);
    if (tree_str[moved] == ':') {
      //      fprintf(stderr, "Proc2 %s\n", tree_str+moved);
      moved += 1;
      moved += read_dist(tree_str+moved, myroot);
      //      fprintf(stderr, "Proc3 %s\n", tree_str+moved);
    }
    else {
      missing_dist = 1;
    }
  }
  else {
    moved = 1;
    myroot->lc = mk_node();
    myroot->lc->par = myroot;
    moved += parse_tree(tree_str+1, myroot->lc);
    if (tree_str[moved] != ',') {
      parse_error(tree_str, moved, ",");
    }
    myroot->rc = mk_node();
    myroot->rc->par = myroot;
    moved +=1;
    moved += parse_tree(tree_str+moved, myroot->rc);
    if (tree_str[moved] != ')') {
      parse_error(tree_str, moved, ")");
    }
    moved += 1;
    if (tree_str[moved] == ':') {
      moved += 1;
      moved += read_dist(tree_str+moved, myroot);
    }
    else {
      missing_dist = 1;
    }
    myroot->leafs = myroot->lc->leafs + myroot->rc->leafs;
  }
  //  fprintf(stderr, " done: %s", tree_str+moved);
  return moved;
}

float get_down_groups (node* mynode, float dsofar, int* numouts) {

  dsofar += (mynode->distup);


  if (mynode->lc || mynode->rc) {
    assert (mynode->lc && mynode->rc);
    do_postorder(mynode->lc); //if postorder has been done this will just quit
    outnames[*numouts] = mynode->lc->seqname;
    outdists[*numouts] = 1/(mynode->lc->distup);
    (*numouts)++;
    do_postorder(mynode->rc);//see above comment
    outnames[*numouts] = mynode->rc->seqname;
    outdists[*numouts] = 1/(mynode->rc->distup);
    (*numouts)++;
    return 1/(mynode->lc->distup) + 1/(mynode->rc->distup);
  }
  else {
    outnames[*numouts] = mynode->seqname;
    outdists[*numouts] = 1/dsofar;
    (*numouts)++;
    return 1;
  }
}

//get me my outgroups
float get_outgroups (node* mynode, float dsofar, int* numouts) {
  node* oc = 0;
  float tsize = 0;
  dsofar += (mynode->distup*(numleafs - mynode->leafs));
  if (mynode->par) {
    if (mynode->par->lc == mynode)
      oc = mynode->par->rc;
    else if (mynode->par->rc == mynode)
      oc = mynode->par->lc;
    else {
      fprintf(stderr, "error in tree structure\n");
      exit(1);
    }
    tsize += get_down_groups(oc, dsofar, numouts);
  }
  return tsize;
}


void do_postorder(node* myroot) {
  int nlen, i, numouts;
  float ttree, test = 0;
  node *fc, *sc;

  //  fprintf(stderr, "into \n");
  if (myroot ->seqname) {// leaf or already done
    //    fprintf(stderr,"if %s\n", myroot->seqname);
    return;
  }
  else {
    //    fprintf(stderr, "else \n");
    do_postorder(myroot->lc);
    do_postorder(myroot->rc);

    nlen = strlen (myroot->lc->seqname) + strlen(myroot->rc->seqname) + 2;
    myroot->seqname = malloc(nlen);
    //    set name    
    if (strcmp(myroot->lc->seqname, myroot->rc->seqname) < 0) {
      fc = myroot->lc;
      sc = myroot->rc;
    }
    else {
      fc = myroot->rc;
      sc = myroot->lc;      
    }

    strcpy(myroot->seqname, fc->seqname);
    strcat(myroot->seqname, "_");
    strcat(myroot->seqname, sc->seqname);

    //    get outgroups
    numouts = 0;
    ttree = get_outgroups(myroot, 0, &numouts);

    fprintf(stdout, "ALIGN %s %s RES %s OUTS", fc->seqname, 
	    sc->seqname, myroot->seqname, ttree);

    for (i=0; i< numouts; i++) {
      if (missing_dist) {
	outdists[i] = 1.0/(float)numouts;
      }
      else {
	outdists[i] = outdists[i]/ttree;
      }
    }
    for (i=0; i< numouts; i++) {
      fprintf(stdout, " %s %f", outnames[i], outdists[i]);
      test += outdists[i];
    }
    fprintf(stdout, "\n");
    //    fprintf(stdout, "%f == %f\n\n", test, ttree);
  }
}

int main(int argc, char** argv) {
  char string_tree[16537]; //noone will ever need more :)))
  int moved;

  node *root = mk_node();
  
  fgets(string_tree, 16537, stdin); 
  if (strlen(string_tree) > 16536) {
    fprintf(stderr, "Wow, that's a biggie... readjust the constants\n");
    return 1;
  }

  moved = parse_tree(string_tree, root);
  if (string_tree[moved] != ';') {
    parse_error(string_tree, moved, ";");
  }
  //  fprintf(stderr, "Parsed tree\n");
  init_consts();
  do_postorder(root);
  return 0;
}
